/*************************************************************************\
    Copyright (c) 2010 Corporation for National Research Initiatives;
                        All rights reserved.
     The CNRI open source license for this software is available at 
                  http://hdl.handle.net/1895.26/1014
---------------------------------------------------------------------------
Developer: Sean Reilly <sreilly@cnri.reston.va.us>
\*************************************************************************/

package net.cnri.apps.doserver.operators;


import net.cnri.apps.doserver.*;
import net.cnri.apps.doserver.txnlog.*;
import net.cnri.do_api.*;
import net.cnri.dobj.*;
import net.cnri.knowbots.lib.*;
import net.cnri.knowbots.station.CollaborationHub;
import net.cnri.util.*;

import org.apache.lucene.analysis.Analyzer;
import org.apache.lucene.analysis.KeywordAnalyzer;
import org.apache.lucene.analysis.PerFieldAnalyzerWrapper;
import org.apache.lucene.analysis.standard.*;
import org.apache.lucene.document.*;
import org.apache.lucene.index.*;
import org.apache.lucene.queryParser.*;
import org.apache.lucene.search.*;
import org.apache.lucene.store.*;
import org.apache.nutch.metadata.*;
import org.apache.nutch.parse.*;
import org.apache.nutch.protocol.*;
import java.io.*;
import java.util.*;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;


/**
 * Operator that provides text indexing capabilities using the Lucene search engine.
 * This version analyzes the transaction log from a particular repository and 
 * indexes any objects in that repository.
 */
public class LuceneIndexer
  extends Knowbot
  implements DOOperation,
             Runnable,
             DOConstants
{
  public static final String DEBUG_PROPERTY = "do.debug_lucene";
  public static final int MAX_INDEXED_BYTES = 500000;
  private static final String INDEX_DIR = "index";
  private static final String DATA_ELEMENT_PREFIX = "lucene.";
  private static final String DEFAULT_MIME_TYPE = "text/plain";
  public static final String LUCENE_CONFIG_ELEMENT = "lucene.config";
  public static final String STATUS_ELEMENT = "lucene.retrieve_status2";
  public static final String FEDERATED_INDEXES_KEY = "federated_indexes";
  
  public static final String DEFAULT_SEARCH_FIELD = "eltext_content";
  private static final String[] SEARCH_FIELDS = {
    "eltext_content",
    "objatt_notes",
    "eltitle_content",
    "elatt_content_filename",
  };
  
  // transaction log constants
  public static final String PARAM_TXN_ID = "txn_id";
  public static final String PARAM_OBJ_ID = "object_id";
  public static final String PARAM_TXN_TYPE = "txn_type";
  public static final String PARAM_TIMESTAMP = "tstamp";
  public static final String PARAM_ACTUAL_TIMESTAMP = "atstamp";
  public static final String PARAM_ELEMENT_ID = "elementid";
  
  public static final String TXN_TYPE_ADD_OBJ = "add";
  public static final String TXN_TYPE_DEL_OBJ = "del";
  public static final String TXN_TYPE_UPDATE_ELEMENT = "update_element";
  public static final String TXN_TYPE_DEL_ELEMENT = "del_element";
  public static final String TXN_TYPE_COMMENT = "comment";
  
  private int NUM_OBJECTS_BATCH_SIZE = 1000; // only index this many objects at a time.
  
  public boolean DEBUG = false;
  
  private static final String[] DEFAULT_ELEMENT_IDS = {
    "metadata", "description", "content"
  };
  
  private static final String LANG_ENGLISH = "en";
  private static final String DEFAULT_LANG = LANG_ENGLISH;
  
  private static final String[] STOP_WORDS_EN =  {
    "a"       , "and"     , "are"     , "as"      ,
    "at"      , "be"      , "but"     , "by"      ,
    "for"     , "if"      , "in"      , "into"    ,
    "is"      , "it"      , "no"      , "not"     ,
    "of"      , "on"      , "or"      , "s"       ,
    "such"    , "t"       , "that"    , "the"     ,
    "their"   , "then"    , "there"   , "these"   ,
    "they"    , "this"    , "to"      , "was"     ,
    "will"    ,
    "with"
  };
  
  
  private transient boolean keepRunning = false; 
  private Hashtable myOperations = new Hashtable();
  private List federatedSearchTargets = new ArrayList();
  private String operationKeys[] = {
    SEARCH_OP_ID
  };
  
  private transient boolean indexIsInitialized = false;
  public boolean enableIndexing = true;
  private DOOperationContext context = null;
  private Main serverMain;
  private DOAuthentication auth = null;
  private DOKeyRing keyring = null;
  private StreamTable indexScanStatus = null;
  private File scanStatusFile = null;
  private File suppressIndexingFile = null;
  private File indexBase = null;
  private StorageProxy repoStorage = null;
  private Analyzer analyzer = null;
  private FSDirectory indexDirectory = null;
  private IndexModifier indexWriter = null;
  private org.apache.hadoop.conf.Configuration nutchConfig = null;
  private HashMap nutchParsers = null;
  private Parser defaultParser = null;
  private Thread shutdownHook = null;
  private Thread indexerThread = null;
  private final TransactionQueueListener queueListener;
  private final ScheduledExecutorService execServ = Executors.newSingleThreadScheduledExecutor();
  private final Object updateWaitCondition = new Object();
  private volatile boolean txnUpdateNeeded = false;
  private volatile boolean periodicUpdateNeeded = true;
  
  public LuceneIndexer(final Main serverMain) 
    throws Exception
  {
    this.serverMain = serverMain;
    
    DEBUG = DEBUG || System.getProperty(DEBUG_PROPERTY, "").toLowerCase().startsWith("y");
    for(int i=0; i<operationKeys.length; i++)
      myOperations.put(operationKeys[i], "");
    auth = AbstractAuthentication.getAnonymousAuth();
    
    
    // add a shutdown hook
    shutdownHook = new Thread() {
      public void run() {
        try {
          keepRunning = false;
          if(queueListener!=null) serverMain.getTxnQueue().removeQueueListener(queueListener);
          if(execServ!=null) execServ.shutdownNow();
          IndexModifier im = indexWriter;
          if(im!=null) im.close();
          saveSourceStatus();
        } catch (Exception e) {
          System.err.println("Error closing indexwriter: "+e);
        }
      }
    };
    Runtime.getRuntime().addShutdownHook(shutdownHook);
    
    
    // load up the indexer configuration from the storage
    initParsers();
    
    String repoID = serverMain.getServerID();
    Storage storage = serverMain.getStorage();
    File serverBase = serverMain.getBaseFolder();
    indexBase = new File(serverBase, INDEX_DIR);
    suppressIndexingFile = new File(indexBase, "suppress_indexing");
    File indexDir = new File(indexBase, "index");
    
    boolean exists = indexDir.exists();
    indexDirectory = FSDirectory.getDirectory(indexDir, !exists);
    PerFieldAnalyzerWrapper analyzerWrapper = new PerFieldAnalyzerWrapper(new StandardAnalyzer(STOP_WORDS_EN));
    analyzerWrapper.addAnalyzer("id",new KeywordAnalyzer());
    analyzerWrapper.addAnalyzer("repoid",new KeywordAnalyzer());
    analyzer = analyzerWrapper;
    if(!exists) {
      IndexWriter writer = new IndexWriter(indexDirectory, analyzer, !exists);
      writer.close();
    }
    
    try {
      indexScanStatus = new StreamTable();
      scanStatusFile = new File(indexBase, "status.dct");
      if(scanStatusFile.exists()) {
        indexScanStatus.readFromFile(scanStatusFile);
      }
    } catch (StringEncodingException e) {
      throw new IOException("Error reading lucene scan status: "+e);
    }

    // kick off the asynchronous indexer..
    indexerThread = new Thread(this);
    indexerThread.setPriority(Thread.MIN_PRIORITY);
    indexerThread.start();

    execServ.scheduleWithFixedDelay(new Runnable() {
        public void run() {
            periodicUpdateNeeded = true;
            synchronized(updateWaitCondition) {
                updateWaitCondition.notify();
            }
        }
    },30,30,TimeUnit.SECONDS);

    queueListener = new LuceneTransactionQueueListener();
    serverMain.getTxnQueue().addQueueListener(queueListener);
  }

  private class LuceneTransactionQueueListener implements TransactionQueueListener {
      public void transactionAdded(Transaction txn) {
          if(!txnUpdateNeeded) {
              txnUpdateNeeded = true;
              synchronized(updateWaitCondition) {
                  updateWaitCondition.notify();
              }
          }
      }
  }
  
  private synchronized void initParsers() {
    // initialize the nutch parser plugins
    try {
      nutchConfig = new org.apache.hadoop.conf.Configuration();
      nutchConfig.setInt("file.content.limit", 65536);
      nutchConfig.set("mime.types.file", "net/cnri/apps/doserver/operators/lucene-mime-types.xml");
      nutchConfig.set("plugin.folders", serverMain.getBaseFolder().getPath() + File.separator + "index" + File.separator + "plugins");
    } catch (Throwable t) {
      t.printStackTrace(System.err);
      System.err.flush();
    }
    
    nutchParsers = new HashMap();
    
    Parser msword = new org.apache.nutch.parse.msword.MSWordParser();
    Parser text = new org.apache.nutch.parse.text.TextParser();
    Parser pdf = new org.apache.nutch.parse.pdf.PdfParser();
    Parser rss = new org.apache.nutch.parse.rss.RSSParser();
    Parser msexcel = new org.apache.nutch.parse.msexcel.MSExcelParser();
    Parser mspp = new org.apache.nutch.parse.mspowerpoint.MSPowerPointParser();
    Parser ootext = new org.apache.nutch.parse.oo.OOParser();
    Parser html = new org.apache.nutch.parse.html.HtmlParser();
    Parser zip = new org.apache.nutch.parse.zip.ZipParser();
    //Parser shockwave = new org.apache.nutch.parse.swf.SWFParser();
    defaultParser = text;
    
    msword.setConf(nutchConfig);
    text.setConf(nutchConfig);
    pdf.setConf(nutchConfig);
    rss.setConf(nutchConfig);
    msexcel.setConf(nutchConfig);
    mspp.setConf(nutchConfig);
    ootext.setConf(nutchConfig);
    try { html.setConf(nutchConfig); } catch (Exception e) {/* grrr */ }
    zip.setConf(nutchConfig);
    
    nutchParsers.put("application/java", text);
    nutchParsers.put("application/pdf", pdf);
    nutchParsers.put("application/postscript", pdf);
    nutchParsers.put("application/msword", msword);
    nutchParsers.put("application/rss+xml", rss);
    nutchParsers.put("application/vnd.ms-excel", msexcel);
    nutchParsers.put("application/vnd.ms-powerpoint", mspp);
    nutchParsers.put("application/vnd.oasis.opendocument.text", ootext);
    nutchParsers.put("application/vnd.oasis.opendocument.text-template", ootext);
    nutchParsers.put("application/vnd.oasis.opendocument.text-master", ootext);
    nutchParsers.put("application/vnd.oasis.opendocument.text-web", ootext);
    nutchParsers.put("application/vnd.oasis.opendocument.presentation", ootext);
    nutchParsers.put("application/vnd.oasis.opendocument.presentation-template", ootext);
    nutchParsers.put("application/vnd.oasis.opendocument.spreadsheet", ootext);
    nutchParsers.put("application/vnd.oasis.opendocument.spreadsheet-template", ootext);
    nutchParsers.put("application/vnd.sun.xml.calc", ootext);
    nutchParsers.put("application/vnd.sun.xml.calc.template", ootext);
    nutchParsers.put("application/vnd.sun.xml.impress", ootext);
    nutchParsers.put("application/vnd.sun.xml.impress.template", ootext);
    nutchParsers.put("application/vnd.sun.xml.writer", ootext);
    nutchParsers.put("application/vnd.sun.xml.writer.template", ootext);
    nutchParsers.put("application/vnd.wap.wbxml", text);
    nutchParsers.put("application/vnd.wap.wmlc", text);
    nutchParsers.put("application/vnd.wap.wmlscriptc", text);
    nutchParsers.put("application/xhtml+xml", text);
    nutchParsers.put("application/x-javascript", "parse-js");
    nutchParsers.put("application/x-kword", msword);
    nutchParsers.put("application/x-kspread", msexcel);
    nutchParsers.put("application/x-latex", text);
    nutchParsers.put("application/x-netcdf", text);
    nutchParsers.put("application/x-sh", text);
    //nutchParsers.put("application/x-shockwave-flash", shockwave);
    nutchParsers.put("application/x-tcl", text);
    nutchParsers.put("application/x-tex", text);
    nutchParsers.put("application/x-texinfo", text);
    nutchParsers.put("application/x-troff", text);
    nutchParsers.put("application/x-troff-man", text);
    nutchParsers.put("application/x-troff-me", text);
    nutchParsers.put("application/x-troff-ms", text);
    nutchParsers.put("application/zip", zip);
    nutchParsers.put("message/news", text);
    nutchParsers.put("message/rfc822", text);
    nutchParsers.put("text/css", text);
    nutchParsers.put("text/html", html);
    nutchParsers.put("text/plain", text);
    nutchParsers.put("text/richtext", msword);
    nutchParsers.put("text/rtf", msword);
    nutchParsers.put("text/sgml", html);
    nutchParsers.put("text/tab-separated-values", msexcel);
    nutchParsers.put("text/vnd.wap.wml", text);
    nutchParsers.put("text/vnd.wap.wmlscript", text);
    nutchParsers.put("text/xml", text);  // html, rss
    nutchParsers.put("text/x-setext", text);
  }
  
  
  public void executeInContext() {
    try {
      indexerThread.join();
    } catch (Exception e) {
      System.err.println("error waiting for indexer thread in LuceneIndexer.executeInContext(): "+e);
      e.printStackTrace(System.err);
    }
  }
  
  
  public boolean canHandleOperation(DOOperationContext context) {
    return myOperations.containsKey(context.getOperationID().toLowerCase());
  }
  
  
  /**
   * Returns a list of operations that this operator can perform
   * on the object identified by the DOOperationContext parameter.
   */
  public String[] listOperations(DOOperationContext context) {
    // we always build an index in the main repository object
    if(context.getTargetObjectID().equals(context.getServerID())) {
      return operationKeys;
    }
    
    try {
      if(context.getStorage().doesDataElementExist(LUCENE_CONFIG_ELEMENT))
        return operationKeys;
    } catch (Exception e) {
      System.err.println("Error checking for element: element="+
                         LUCENE_CONFIG_ELEMENT+" error="+e);
    }
    
    return null;
  }
  
  
  public void run() {
    keepRunning = true;
    System.err.println("Starting Lucene Indexer async indexing");
    
    try {
      while(keepRunning) {
        boolean bigUpdate = periodicUpdateNeeded;
        try {
            updateIndex(bigUpdate);
        } catch (Exception t) {
            System.err.println("Error updating index: "+t);
            t.printStackTrace(System.err);
        }
        if(bigUpdate) periodicUpdateNeeded = false;
        
        synchronized(updateWaitCondition) {
            while(!txnUpdateNeeded && !periodicUpdateNeeded) {
                try {
                    updateWaitCondition.wait();
                }
                catch(InterruptedException e) {
                    break;
                }
            }
        }
      }
    } finally {
      Thread shutdown = this.shutdownHook;
      if(shutdown!=null) {
        Runtime.getRuntime().removeShutdownHook(shutdownHook);
        shutdown.run();
        this.shutdownHook = null;
      }
    }
  }
  
  
  /**
   * Performs the given operation (which this object has advertised that it
   * can handle) which consists of reading input (if any is expected) from the
   * given InputStream and writing the output of the operation (if any) to the
   * OutputStream.  This method should *always* close the input and output streams
   * when finished with them.  If there are any errors in the input, the error
   * message must be communicated on the OutputStream since all errors must be
   * at the application level.  Any exceptions thrown by this method will *not*
   * be communicated to the caller and are therefore not acceptable.
   */
  public void performOperation(DOOperationContext context, InputStream in, OutputStream out) {
    try {
      String operation = context.getOperationID();
      if(operation.equalsIgnoreCase(SEARCH_OP_ID)) {
        doSearch(context, in, out);
      } else {
        // shouldn't happen
        sendErrorResponse(out, "Operation '"+operation+"' not implemented!",
                          DOException.OPERATION_NOT_AVAILABLE);
        return;
      }
    } catch (Exception e) {
      if(DEBUG) e.printStackTrace(System.err);
    } finally {
      try { in.close(); } catch (Exception e) {}
      try { out.close(); } catch (Exception e) {}
    }
  }
  
  private MultiFieldQueryParser searchParser = null;
  private final Object searchParserLock = new Object();
  
  /**
   * Implements the search operation.  This performs a search on the index
   * in the current object based on the 'query' string parameter.
   */
  private void doSearch(DOOperationContext context, InputStream in, OutputStream out)
    throws Exception
  {
    StorageProxy storage = context.getStorage();
    HeaderSet params = context.getOperationHeaders();
    String queryStr = params.getStringHeader("query", null);
    
    String queryLang = params.getStringHeader("language", DEFAULT_LANG).toLowerCase();
    if(queryStr==null) {
      sendErrorResponse(out, "Request was missing 'query' parameter",
                        DOException.APPLICATION_ERROR);
      return;
    }
    if(analyzer==null) {
      sendErrorResponse(out, "Unrecognized language: "+queryLang,
                        DOException.APPLICATION_ERROR);
      return;
    }
    
    if(DEBUG) System.err.println("searching lucene for: "+queryStr);
    
    Query q0;
    synchronized(searchParserLock) {
        if(searchParser==null) {
            searchParser = new MultiFieldQueryParser(SEARCH_FIELDS, analyzer);
            searchParser.setDefaultOperator(QueryParser.AND_OPERATOR);
        }
    
        // QueryParser is NOT thread-safe.  I found this out the hard way
        q0 = searchParser.parse(queryStr);
    }
    
    //q0.setBoost(10.0f);
    //BooleanQuery query = new BooleanQuery();
    //query.add(q0, BooleanClause.Occur.MUST);
    
    IndexReader reader = IndexReader.open(indexDirectory);
    if(DEBUG) System.err.println("search for <<<"+q0+">>> index-size:"+reader.numDocs());
    IndexSearcher indexSearcher = new IndexSearcher(reader);
    
    Hits hits = indexSearcher.search(q0);
    if(DEBUG) System.err.println("Found "+hits.length()+" results.");
    sendSuccessResponse(out);

    ArrayList federatedThreads = new ArrayList();
    boolean doFederation = federatedSearchTargets!=null && federatedSearchTargets.size()>0 && 
      params.getBooleanHeader("federation", true);
    
    if(doFederation) {
      // perform federated search, returning results from searches on other repositories
      params.removeHeadersWithKey("federation");
      params.addHeader("federation", false);
      
      for(Iterator iter=federatedSearchTargets.iterator(); iter.hasNext(); ) {
        FederatedSearcher fedSearch = new FederatedSearcher((String)iter.next(), context, params, out);
        fedSearch.start();
        federatedThreads.add(fedSearch);
      }
    }
    
    net.cnri.apps.doserver.Authorizer authorizer = serverMain.getAuthorizer();
    
    String tmp = null;
    HeaderSet result = new HeaderSet("result");
    for (Iterator it=hits.iterator(); it.hasNext(); ) {
      result.removeAllHeaders();
      Hit hit = (Hit)it.next();
      Document doc = hit.getDocument();
      String objectID = doc.get("id");
      if(objectID==null) continue;
      
      // only return this result if the caller has read access to the object:
      if(authorizer.operationIsAllowed(context, objectID, DOConstants.GET_DATA_OP_ID)) {
        result.addHeader("objectid", objectID);
        result.addHeader("repoid", serverMain.getServerID());
        result.addHeader("score", String.valueOf(hit.getScore()));
        for(Enumeration en=doc.fields(); en.hasMoreElements(); ) {
          Field f = (Field)en.nextElement();
          String val = f.stringValue();
          if(val==null) val = "";
          result.addHeader("field:"+f.name(), val);
        }
        synchronized(out) {
          result.writeHeaders(out);
        }
      }
    }
    
    // wait for the federated searches to finish
    for(Iterator iter=federatedThreads.iterator(); iter.hasNext(); ) {
      ((Thread)iter.next()).join();
    }
    
    out.close();
  }
  
  
  private class FederatedSearcher
    extends Thread
  {
    private String searchTarget;
    private DOOperationContext context;
    private HeaderSet params;
    private OutputStream out;
    
    FederatedSearcher(String searchTarget, DOOperationContext context, HeaderSet params,
                      OutputStream out) {
      this.searchTarget = searchTarget;
      this.context = context;
      this.params = params;
      this.out = out;
    }
    
    public void run() {
      try {
        if(DEBUG) System.err.println("federating search to "+searchTarget);
        context.performOperation(null, searchTarget, SEARCH_OP_ID, params,
                                 new ByteArrayInputStream(new byte[0]), 
                                 new DontCloseOutputStream(this, out));
        if(DEBUG) System.err.println("finished search federation to "+searchTarget);
      } catch (Exception e) {
        System.err.println("Error federating search to index: "+searchTarget+": "+e);
        e.printStackTrace(System.err);
      }
    }
    
    public String toString() {
      return "fedsearch at '"+searchTarget+"' for terms '"+params+"'";
    }
    
    void receivedResult(byte line[]) 
      throws IOException
    {
      synchronized(out) { out.write(line); }
      if(DEBUG) System.err.println("result from "+searchTarget+":  "+new String(line));
    }
  }
  
  
  private class DontCloseOutputStream
  extends OutputStream
  {
    private OutputStream out;
    private ByteArrayOutputStream bout;
    private boolean sawStatusLine = false;
    private FederatedSearcher searcher;
    
    DontCloseOutputStream(FederatedSearcher searcher, OutputStream out) {
      this.searcher = searcher;
      this.out = out;
      this.bout = new ByteArrayOutputStream();
    }
    
    public void write(int b) 
    throws IOException
    {
      bout.write(b);
      if(((byte)b)=='\n') {
        byte line[] = bout.toByteArray();
        // TODO: Check for first line, which should be the operation's status and skip it
        if(!sawStatusLine) {
          if(DEBUG) System.err.println("got status from search: "+net.handle.hdllib.Util.decodeString(line));
          sawStatusLine = true;
        } else {
          searcher.receivedResult(line);
        }
        bout.reset();
      }
    }
    
    public void close() {
      // noop
    }
  }
  
  
  /** Updates the indexes from any sources specified in the config file */
  private void updateIndex(boolean bigUpdate) {
    boolean txnUpdateStillNeeded = txnUpdateNeeded;
    StreamTable indexConfig = null;
    try {
        // check for the existence of the "suppress_indexing" file
        if(suppressIndexingFile.exists()) return;

        if(bigUpdate) {
            // (re-)initialize the authentication used for indexing
            CollaborationHub hub = serverMain.getServiceRegistry();
            Object keyrings[] = hub.lookupEntity(null, "lucene-indexing-keyring", DOKeyRing.class);
            if(keyrings!=null && keyrings.length>0) {
                this.keyring = (DOKeyRing)keyrings[0];
            } else {
                this.keyring = null;
            }
        }

        try {
            if(indexWriter==null) {
                indexWriter = new IndexModifier(indexDirectory, analyzer, false);
            }
        } catch (IOException e) {
            System.err.println("Error opening index writer: "+e);
            e.printStackTrace(System.err);
            return;
        }

        if(bigUpdate) {
            InputStream in = null;
            indexConfig = new StreamTable();
            try {
                if(repoStorage==null) {
                    repoStorage = new ConcreteStorageProxy(serverMain.getStorage(), 
                            serverMain.getServerID(),
                            serverMain.getServerID(),
                            null);
                }
                if(repoStorage.doesDataElementExist(LUCENE_CONFIG_ELEMENT)) {
                    in = repoStorage.getDataElement(LUCENE_CONFIG_ELEMENT);
                }
                if(in!=null) {
                    indexConfig.readFrom(in);

                    Object fedVect = indexConfig.get(FEDERATED_INDEXES_KEY);
                    if(fedVect!=null && fedVect instanceof List) {
                        federatedSearchTargets = (List)fedVect;
                    } else {
                        federatedSearchTargets = new ArrayList();
                    }
                } else {
                    federatedSearchTargets = new ArrayList();
                }
            } catch (Exception e) {
                System.err.println("Error reading configuration: "+e);
                e.printStackTrace(System.err);
                return;
            } finally {
                try { in.close(); } catch (Exception e) {}
            }
        }    
        txnUpdateNeeded = false;
        txnUpdateStillNeeded = false;
    }
    finally {
        if(txnUpdateStillNeeded) {
            try {
                Thread.sleep(30 * 1000);
            }
            catch(InterruptedException e) {
                Thread.currentThread().interrupt();
            }
        }
    }
    
    try {
      // retrieve any updates from the repository in which we live
      updateFromContainer();
      
      if(bigUpdate) {
          // retrieve any updates from other repositories
          StreamTable config = (StreamTable)indexConfig.deepClone();
          StreamVector sources = (StreamVector)config.get("sources");
          for(int i=0; sources!=null && i<sources.size(); i++) {
              updateFromSource(String.valueOf(sources.get(i)));
          }

          // scan objects that other indexes couldn't scan in the hopes that we 
          // have the necessary rights to access and decrypt those objects
          StreamVector sourceIndexes = (StreamVector)indexConfig.get("srcindexes");
          for(int i=0; sourceIndexes!=null && i<sourceIndexes.size(); i++) {
              updateFromIndex(String.valueOf(sourceIndexes.get(i)));
          }

          try {
              indexWriter.optimize();
          } catch (IOException e) {
              System.err.println("Error optimizing index: "+e);
          }
      }
    } finally {
      try {
        if(indexWriter!=null) {
          indexWriter.close();
          indexWriter = null;
        }
      } catch (Exception e) {
        System.err.println("Error closing index after update: "+e);
        e.printStackTrace(System.err);
      }
    }
    
    
  }
  
  
  
  /** Updates the index from the given source */
  private void updateFromIndex(String indexID) {
    if(DEBUG) System.err.println("Updating from index: "+indexID);
    StreamTable status = (StreamTable)indexScanStatus.get("index:"+indexID);
    synchronized(indexScanStatus) {
      if(status==null) {
        status = new StreamTable();
        indexScanStatus.put("index:"+indexID, status);
      }
    }
    DOClient client = new DOClient(auth);
    HashMap repos = new HashMap();
    
    HeaderSet params = new HeaderSet();
    String lastTimeStamp = status.getStr("last_idx_timestamp", null);
    StringBuffer query = new StringBuffer();
    query.append("idxerror:").append(DOException.PERMISSION_DENIED_ERROR).append(' ');
    if(lastTimeStamp!=null) {
      query.append("dtindexed:[").append(lastTimeStamp).append(" TO 999999999999999999999] ");
    }
    params.addHeader("query", query.toString());
    
    DOAuthentication localAuth = auth;
    if(localAuth==null) {
      localAuth = PKAuthentication.getAnonymousAuth();
    }
    
    StreamPair io = null;
    try {
      io = client.performOperation(indexID, SEARCH_OP_ID, params);
      HeaderSet response = new HeaderSet();
      InputStream in = io.getInputStream();
      
      while(response.readHeaders(in)) {
        String objectID = response.getStringHeader("objectid", null);
        if(objectID==null) continue;
        String repoID = response.getStringHeader("repoid", null);
        System.err.println("indexing restricted record: "+objectID+" @ "+repoID);
        
        if(repoID==null) {
          repoID = DOClient.resolveRepositoryID(objectID);
        }
        
        Repository repo = (Repository)repos.get(repoID);
        if(repo==null) {
          repo = new Repository(localAuth, repoID);
          repos.put(repoID, repo);
        }
        
        updateObjectInIndex(repo, objectID);
        
        String lastIndexed = response.getStringHeader("field:dtindexed", null);
        if(lastIndexed!=null) {
          status.put("last_idx_timestamp", lastIndexed.trim());
        }
      }
    } catch (Exception e) {
      System.err.println("Error updating from index "+indexID+": "+e);
      e.printStackTrace(System.err);
    } finally {
      try { io.close(); } catch (Throwable t) {}
      saveSourceStatus();  // save the source/index status info
    }
    
  }
  
  
  
  
  private void updateFromContainer() {
    HashMap updatedObjects = new HashMap();
    
    try {
      long lastTimestamp = indexScanStatus.getLong("self_scan_txn", -1);
      ArrayList objectsToUpdate = new ArrayList();
      
      if(lastTimestamp==-1) {
        // we haven't done an update yet, scan all the objects in the repository
        // scan all of the objects...
        Enumeration allObjects = repoStorage.listObjects();
        while(allObjects.hasMoreElements()) {
          updateObjectFromStorage((String)allObjects.nextElement());
        }
        
        lastTimestamp = 0;
        // get the most recent timestamp, unfortunately by scanning the entire queue (at least it is local)
        Enumeration txnEnum = serverMain.getTxnQueue().getScanner(lastTimestamp);
        while(txnEnum.hasMoreElements()) {
          Transaction txn = (Transaction)txnEnum.nextElement();
          lastTimestamp = txn.timestamp;
        }
      } else {
        // we are updating all objects since the last transaction update
        Enumeration txnEnum = serverMain.getTxnQueue().getScanner(lastTimestamp);
        while(txnEnum.hasMoreElements()) {
          Transaction txn = (Transaction)txnEnum.nextElement();
          
          if(txn.objectID==null) continue; // ignore all non-object specific transactions
          
          switch(txn.action) {
            case Transaction.ACTION_COMMENT:
              break;
            case Transaction.ACTION_OBJ_DEL:
            case Transaction.ACTION_OBJ_ADD:
            case Transaction.ACTION_OBJ_DATA_UPDATE:
            case Transaction.ACTION_OBJ_DATA_DELETE:
            case Transaction.ACTION_OBJ_ATTRIBUTE_UPDATE:
            case Transaction.ACTION_OBJ_ATTRIBUTE_DELETE:
              if(!updatedObjects.containsKey(txn.objectID)) {
                // only update the object once per session
                objectsToUpdate.add(txn.objectID);
                updatedObjects.put(txn.objectID, "");
              }
              break;
          }
          
          lastTimestamp = txn.timestamp;
          
          if(objectsToUpdate.size() >= NUM_OBJECTS_BATCH_SIZE) {
            break;
          }
        }
        
        for(Iterator iter=objectsToUpdate.iterator(); iter.hasNext(); ) {
          String objectID = (String)iter.next();
          if (DEBUG) System.err.println("updating object: " + objectID + " from storage");
          updateObjectFromStorage(objectID);
        }
      }
      
      indexScanStatus.put("self_scan_txn", lastTimestamp);
    } catch (Exception e) {
      System.err.println("Error trying to get txns from storage: "+e);
      if(DEBUG) e.printStackTrace(System.err);
    } finally {
      saveSourceStatus();  // save the source/index status info
    }
  }
  
  
  
  /** Updates the index from the given source */
  private void updateFromSource(String sourceID) {
    StreamTable sourceStatus;
    
    DOClientConnection connection = null;
    DOServiceInfo svcInfo = null;
    if(DEBUG) System.err.println("updating from "+sourceID);
    try {
      svcInfo = new DOServiceInfo(sourceID);
    } catch (Exception e) {
      System.err.println("Error resolving source: "+sourceID+"; error: "+e);
      return;
    }
    
    synchronized(indexScanStatus) {
      sourceStatus = (StreamTable)indexScanStatus.get(sourceID);
      if(sourceStatus==null) {
        indexScanStatus.put(sourceID, sourceStatus = new StreamTable());
      }
    }
    
    
    for(int i=0; i<svcInfo.getServerCount(); i++) {
      DOServerInfo svrInfo = svcInfo.getServer(i);
      if(svrInfo==null) continue;
      
      String serverID = svrInfo.getServerID();
      StreamTable serverStatus = null;
      synchronized(sourceStatus) {
        serverStatus = (StreamTable)sourceStatus.get(serverID);
        if(serverStatus==null) {
          sourceStatus.put(serverID, serverStatus = new StreamTable());
        }
      }
      
      if(DEBUG) System.err.println("updating from: src="+sourceID+"; svr="+serverID);
      HeaderSet params = new HeaderSet();
      if(serverStatus.containsKey("last_txn_id")) {
        params.addHeader("txn_id", serverStatus.getStr("last_txn_id", null));
      }
      
      StreamPair io = null;
      Repository repo = null;
      DOClientConnection conn = null;
      try {
        DOAuthentication localAuth = auth;
        if(localAuth==null) {
          localAuth = PKAuthentication.getAnonymousAuth();
        }
        conn = new DOClientConnection(localAuth);
        conn.setUseEncryption(false);
        conn.connect(svrInfo);
        
        repo = new Repository(sourceID, conn);
        
        io = conn.performOperation(sourceID, DOConstants.GET_REPO_TXNS_OP_ID,
                                   params);
        
        InputStream in = io.getInputStream();
        HeaderSet record = new HeaderSet();
        HashMap updatedObjects = new HashMap();
        
        while(record.readHeaders(in)) {
          String timestampStr = record.getStringHeader("tstamp", null);
          if(timestampStr==null) {
            System.err.println("Warning:  ignoring received transaction with "+
                               "null timestamp: "+record);
            continue;
          }
          
          String objID = record.getStringHeader("object_id", null);
          if(objID==null) continue; // ignore all non-object specific transactions
          long timestamp = record.getLongHeader("tstamp", 0);
          
          String txnType = record.getStringHeader(PARAM_TXN_TYPE, "");
          if(txnType.equals(TXN_TYPE_COMMENT)) {
            // nothing to index...
          } else if(txnType.equals(TXN_TYPE_DEL_OBJ)) {
            // delete the object from the index
            deleteObjectFromIndex(objID);
          } else {
            // for all other modifications update the object in the index
            
            if(!updatedObjects.containsKey(objID)) {
              // only update the object once per session
              updateObjectInIndex(repo, objID);
              updatedObjects.put(objID, "");
            }
          }
          
          serverStatus.put("last_txn_id", timestamp);
        }
      } catch (Exception e) {
        System.err.println("Error trying to get txns from src="+sourceID+
                           " svr="+serverID+": "+e);
        if(DEBUG) e.printStackTrace(System.err);
      } finally {
        try { io.close(); } catch (Throwable t) {}
        saveSourceStatus();  // save the source status info
        try {
          conn.close();
        } catch (Throwable t) {
          System.err.println("Warning: Error closing connection to "+sourceID+": "+t);
        }
        conn = null;
      }
    }
  }
  
  
  private void saveSourceStatus() {
    try {
      synchronized(indexScanStatus) {
        indexScanStatus.writeToFile(scanStatusFile);
      }
    } catch (Exception e) {
      System.err.println("Error saving replication status: "+e);
      if(DEBUG) e.printStackTrace(System.err);
    }
  }
  
  /** Re-retrieve the object and add it to the index */
  private void deleteObjectFromIndex(String objID) 
    throws IOException
  {
    if(DEBUG) System.err.println("deleting object: "+objID);
    
    indexWriter.deleteDocuments(new Term("id", objID));
    indexWriter.flush();
  }
  
  
  /** Update the object directly from the storage module */
  private void updateObjectFromStorage(String objectID) 
  throws Exception
  {
    if(DEBUG) System.err.println("Indexing object from storage: "+objectID);
    Thread thread = Thread.currentThread();
    String oldThreadName = thread.getName();
    thread.setName("Indexing object '"+objectID+"' directly from storage");
    try {
      StorageProxy objStore = repoStorage.getObjectAccessor(objectID);
      if(!objStore.doesObjectExist()) {
        indexWriter.deleteDocuments(new Term("id", objectID)); // delete any old entries for this ID
        indexWriter.flush();
        return;
      }
      Document doc = new Document();
      doc.add(new Field("id", objectID, Field.Store.YES, Field.Index.UN_TOKENIZED));
      
      doc.add(new Field("repoid", serverMain.getServerID(), Field.Store.YES,
                        Field.Index.UN_TOKENIZED));
      doc.add(new Field("indextime", String.valueOf(System.currentTimeMillis()),
                        Field.Store.YES, Field.Index.UN_TOKENIZED));
      
      // store the object attributes in the index
      HeaderSet attributes = objStore.getAttributes(null);
      for(Iterator iter=attributes.iterator(); iter.hasNext(); ) {
        HeaderItem attribute = (HeaderItem)iter.next();
        doc.add(new Field("objatt_"+attribute.getName().replace(':','_'),
                          attribute.getValue(),
                          Field.Store.YES, Field.Index.TOKENIZED));
      }
      
      doc.add(new Field("objcreated", 
                        DateTools.dateToString(new java.util.Date(attributes.getLongHeader(DATE_CREATED_ATTRIBUTE, 0)),
                                               DateTools.Resolution.MILLISECOND),
                        Field.Store.YES, Field.Index.UN_TOKENIZED));
      doc.add(new Field("objmodified", 
                        DateTools.dateToString(new java.util.Date(attributes.getLongHeader(DATE_MODIFIED_ATTRIBUTE, 0)),
                                               DateTools.Resolution.MILLISECOND),
                        Field.Store.YES, Field.Index.UN_TOKENIZED));
      
      for(Enumeration en=objStore.listDataElements(); en.hasMoreElements(); ) {
        String elementID = (String)en.nextElement();
        addFieldForElement(doc, objectID, objStore, elementID);
      }
      
      // if the internal.rights element exists, add a compiled-down form to the index
      if(objStore.doesDataElementExist(DOConstants.RIGHTS_ELEMENT_ID)) {
        net.cnri.apps.doserver.Authorizer authorizer = serverMain.getAuthorizer();
        String[] readers = authorizer.getExplicitAccepts(objStore, DOConstants.GET_DATA_OP_ID);
        StringBuffer readerStr = new StringBuffer();
        for(int i=0; readers!=null && i<readers.length; i++) {
          if(i!=0) readerStr.append('\n');
          readerStr.append(readers[i].replaceAll("\n", "\\n"));
        }
        doc.add(new Field("readers", readerStr.toString(), Field.Store.YES, Field.Index.TOKENIZED));
        String[] writers = authorizer.getExplicitAccepts(objStore, DOConstants.STORE_DATA_OP_ID);
        StringBuffer writerStr = new StringBuffer();
        for(int i=0; writers!=null && i<writers.length; i++) {
          if(i!=0) writerStr.append('\n');
          writerStr.append(writers[i].replaceAll("\n", "\\n"));
        }
        doc.add(new Field("writers", writerStr.toString(), Field.Store.YES, Field.Index.TOKENIZED));
      }
      
      
      //if(DEBUG) System.err.println("storing "+objID+" in index as doc:"+doc);
      doc.add(new Field("dtindexed",
                        DateTools.dateToString(new Date(),
                                               DateTools.Resolution.MILLISECOND),
                        Field.Store.YES, Field.Index.UN_TOKENIZED));
      indexWriter.deleteDocuments(new Term("id", objectID)); // delete any old entries for this ID
      indexWriter.addDocument(doc);  // add the new document
      indexWriter.flush();
    } finally {
      thread.setName(oldThreadName);
    }
  }

  /** Re-retrieve the object and add it to the index */
  private void updateObjectInIndex(Repository repo, String objectID) 
  throws IOException
  {
    StreamPair io = null;
    Document doc = new Document();
    doc.add(new Field("id", objectID, Field.Store.YES, Field.Index.UN_TOKENIZED));
    doc.add(new Field("repoid", repo.getID(), Field.Store.YES,
                      Field.Index.UN_TOKENIZED));
    doc.add(new Field("indextime", String.valueOf(System.currentTimeMillis()),
                      Field.Store.YES, Field.Index.UN_TOKENIZED));
    try {
      DigitalObject obj = repo.getDigitalObject(objectID);
      String[] elementIDs = obj.listDataElements();
      String attributes[] = obj.listAttributes();
      for(int i=0; attributes!=null && i<attributes.length; i++) {
        String attName = attributes[i];
        doc.add(new Field("objatt_"+attName.replace(':','_'),
                          obj.getAttribute(attName, ""),
                          Field.Store.YES, Field.Index.TOKENIZED));
      }
      doc.add(new Field("objcreated", 
                        DateTools.dateToString(obj.getDateCreated(),
                                               DateTools.Resolution.MILLISECOND),
                        Field.Store.YES, Field.Index.UN_TOKENIZED));
      doc.add(new Field("objmodified", 
                        DateTools.dateToString(obj.getDateLastModified(),
                                               DateTools.Resolution.MILLISECOND),
                        Field.Store.YES, Field.Index.UN_TOKENIZED));
      
      // add the different attributes to the document...
      for(int i=0; i<elementIDs.length; i++) {
        //if(!elementIDs[i].startsWith("content")) continue;
        addFieldForElement(doc, obj.getDataElement(elementIDs[i]));
      }
    } catch (DOException e) {
      System.err.println("Error indexing object "+objectID+" in "+repo+": "+e);
      if(e.getErrorCode()!=DOException.PERMISSION_DENIED_ERROR) {
        return;
      }
      doc.add(new Field("idxerror", String.valueOf(e.getErrorCode()),
                        Field.Store.YES, Field.Index.UN_TOKENIZED));
    }
    
    //if(DEBUG) System.err.println("storing "+objID+" in index as doc:"+doc);
    doc.add(new Field("dtindexed",
                      DateTools.dateToString(new Date(),
                                             DateTools.Resolution.MILLISECOND),
                      Field.Store.YES, Field.Index.UN_TOKENIZED));
    
    indexWriter.deleteDocuments(new Term("id", objectID)); // delete any old entries for this ID
    indexWriter.addDocument(doc);  // add the new document
    indexWriter.flush();
  }
  
  
  private InputStream getElement(DOClientConnection conn,
                                 String objID, String elementID) 
    throws DOException, IOException
  {
    HeaderSet params = new HeaderSet();
    params.addHeader("elementid", elementID);
    StreamPair io = conn.performOperation(objID, DOConstants.GET_DATA_OP_ID, params);
    io.getOutputStream().close();
    return io.getInputStream();
  }
  
  
  
  /** Index the given data element and add the contents to the given lucene document. */
  private void addFieldForElement(Document doc, String objectID, StorageProxy objStore, String elementID) {
    if(DEBUG) System.err.println("  element: "+elementID);
    InputStream in = null;
    try {
      String defaultMimeType = null;
      if(elementID.equals(DOConstants.RIGHTS_ELEMENT_ID)) {
        defaultMimeType = "text/plain";
      }
      
      HeaderSet attributes = objStore.getElementAttributes(elementID, null);
      String mimeType = attributes.getStringHeader(DOConstants.MIME_TYPE_ATTRIBUTE, null);
      String title = attributes.getStringHeader(DOConstants.TITLE_ATTRIBUTE, null);
      String elementText = null;
      boolean encrypted = attributes.getStringHeader(DOKeyRing.KEY_ID_ATTRIBUTE, null)!=null;
      if(encrypted) {
        doc.add(new Field("has_enc_el", "1", Field.Store.YES, Field.Index.UN_TOKENIZED));
      }
      Metadata md = new Metadata();
      Parser parser = null;
      if(mimeType!=null && (!encrypted || keyring!=null)) {
        if(mimeType.indexOf(';')>=0) {
          mimeType = mimeType.substring(0, mimeType.indexOf(';')).trim();
        }
        md.set(Metadata.CONTENT_TYPE, mimeType);
        parser = (Parser)nutchParsers.get(mimeType.toLowerCase());
        
        // if we have a parser, retrieve the content
        if(parser!=null) {
          in = objStore.getDataElement(elementID);
          
          // read the contents of the element
          ByteArrayOutputStream bout = new ByteArrayOutputStream();
          byte buf[] = new byte[2048];
          int r;
          int numRead = 0;
          while((r=in.read(buf))>=0) {
            bout.write(buf, 0, r);
            numRead += r;
          }
          try { in.close(); } catch (Throwable t) {}
          md.set(Metadata.CONTENT_LENGTH, String.valueOf(numRead));
          
          byte contentBytes[] = bout.toByteArray();
          Content elementContent = new Content(elementID, "hdl:"+objectID,
                                               contentBytes, mimeType, md, nutchConfig);
          try {
            Parse parse = parser.getParse(elementContent);
            ParseData parseData = parse.getData();
            Metadata parseMD = parseData.getContentMeta();
            elementText = parse.getText();
            
            if(title==null) title = parseData.getTitle();
            String mdNames[] = parseMD.names();
            for(int i=0; mdNames!=null && i<mdNames.length; i++) {
              doc.add(new Field("elpatt_"+elementID.replace(':','_')+"_"+mdNames[i].replace(':','_'), 
                                parseMD.get(mdNames[i]),
                                Field.Store.YES, Field.Index.UN_TOKENIZED));
            }
            Metadata parseMD2 = parseData.getParseMeta();
            mdNames = parseMD2.names();
            for(int i=0; mdNames!=null && i<mdNames.length; i++) {
              doc.add(new Field("elmpatt_"+elementID.replace(':','_')+"_"+mdNames[i].replace(':','_'), 
                                parseMD2.get(mdNames[i]),
                                Field.Store.YES, Field.Index.UN_TOKENIZED));
            }
          } catch (RuntimeException e) {
            System.err.println("runtime parse error: "+e);
          } catch (Exception e) {
            System.err.println("parse error: "+e);
          }
        }
      }
      
      String filename = attributes.getStringHeader(DOConstants.FILE_NAME_ATTRIBUTE, null);
      if((title==null || title.length()<=0) && filename!=null && filename.trim().length()>0) {
        title = filename;
      }
      
      if((title==null || title.length()<=0) && elementText!=null) {
        // try to pick out a decent title
        title = elementText;
        if(title.length()>200) {
          title = title.substring(0, 200).trim();
          int nlIdx = title.indexOf('\n');
          if(nlIdx>0) title = title.substring(0, nlIdx).trim();
        } else {
          title = elementText.trim();
        }
      }
      
      if(DEBUG) {
        String textToPrint = null;
        if(elementText!=null) {
          textToPrint = elementText.trim();
          if(textToPrint.length()>100)
            textToPrint = textToPrint.substring(0, 100).trim();
        }
        System.err.println("  title: "+title);
        System.err.println("  mimetype: "+mimeType);
        System.err.println("  text: "+textToPrint);
      }
      
      if(elementText==null || elementText.trim().length()<=0) {
        // the parser failed
        System.err.println("Warning: no text parsed from element "+elementID+" in object "+objectID);
      }
      
      if(title!=null) {
        doc.add(new Field("eltitle_"+elementID.replace(':','_'), title,
                          Field.Store.YES, Field.Index.TOKENIZED));
      }
      if(mimeType!=null) {
        doc.add(new Field("eltype_"+elementID.replace(':','_'), mimeType,
                          Field.Store.YES, Field.Index.UN_TOKENIZED));
      }
      if(elementText!=null) {
        doc.add(new Field("eltext_"+elementID.replace(':','_'), elementText,
                          Field.Store.NO, Field.Index.TOKENIZED));
      }
      for(Iterator iter=attributes.iterator(); iter.hasNext(); ) {
        HeaderItem attribute = (HeaderItem)iter.next();
        doc.add(new Field("elatt_"+elementID.replace(':','_')+"_"+attribute.getName().replace(':','_'), 
                          attribute.getValue(),
                          Field.Store.YES, Field.Index.TOKENIZED));
      }
    } catch (Exception e) {
      System.err.println("error getting element "+elementID+" from object "+objectID+": "+e);
      if(DEBUG) {
        e.printStackTrace(System.err);
        System.err.flush();
      }
    } finally {
      try { if(in!=null) in.close(); } catch (Throwable t) {}
    }
    
    
  }
  
  
  /** Initiate a retrieval of the given data element and add the contents to the
    * given lucene document.  Doesn't close the Reader that is passed as the Field
    * that is added to the Document. */
  private void addFieldForElement(Document doc, DataElement element) {
    if(DEBUG) System.err.println("  element: "+element);
    String elementID = element.getDataElementID();
    InputStream in = null;
    try {
      String mimeType = element.getAttribute(DOConstants.MIME_TYPE_ATTRIBUTE, null);
      String title = element.getAttribute(DOConstants.TITLE_ATTRIBUTE, null);
      String elementText = null;
      boolean encrypted = element.getAttribute(DOKeyRing.KEY_ID_ATTRIBUTE, null)!=null;
      if(encrypted) {
        doc.add(new Field("has_enc_el", "1", Field.Store.YES, Field.Index.UN_TOKENIZED));
      }
      Metadata md = new Metadata();
      Parser parser = null;
      if(mimeType!=null && (!encrypted || keyring!=null)) {
        if(mimeType.indexOf(';')>=0) {
          mimeType = mimeType.substring(0, mimeType.indexOf(';')).trim();
        }
        md.set(Metadata.CONTENT_TYPE, mimeType);
        parser = (Parser)nutchParsers.get(mimeType.toLowerCase());
        
        // if we have a parser, retrieve the content
        if(parser!=null) {
          if(encrypted && keyring!=null) {
            try {
              in = keyring.decryptDataElement(element);
            } catch (Exception e) {
              System.err.println("Error decrypting element: "+e);
              in = element.read();
            }
          } else {
            in = element.read();
          }
          
          // read the contents of the element
          
          ByteArrayOutputStream bout = new ByteArrayOutputStream();
          byte buf[] = new byte[2048];
          int r;
          int numRead = 0;
          while((r=in.read(buf))>=0) {
            bout.write(buf, 0, r);
            numRead += r;
          }
          try { in.close(); } catch (Throwable t) {}
          md.set(Metadata.CONTENT_LENGTH, String.valueOf(numRead));
          
          byte contentBytes[] = bout.toByteArray();
          Content elementContent = 
            new Content(elementID, "hdl:"+element.getDigitalObject().getID(),
                        contentBytes, mimeType, md, nutchConfig);
          
          Parse parse = parser.getParse(elementContent);
          ParseData parseData = parse.getData();
          Metadata parseMD = parseData.getContentMeta();
          elementText = parse.getText();
          
          if(title==null) title = parseData.getTitle();
          String mdNames[] = parseMD.names();
          for(int i=0; mdNames!=null && i<mdNames.length; i++) {
            doc.add(new Field("elpatt_"+elementID.replace(':','_')+"_"+mdNames[i].replace(':','_'), 
                              parseMD.get(mdNames[i]),
                              Field.Store.YES, Field.Index.UN_TOKENIZED));
          }
          Metadata parseMD2 = parseData.getParseMeta();
          mdNames = parseMD2.names();
          for(int i=0; mdNames!=null && i<mdNames.length; i++) {
            doc.add(new Field("elmpatt_"+elementID.replace(':','_')+"_"+mdNames[i].replace(':','_'), 
                              parseMD2.get(mdNames[i]),
                              Field.Store.YES, Field.Index.UN_TOKENIZED));
          }
        }
      }
      
      if((title==null || title.length()<=0) && elementText!=null) {
        // try to pick out a decent title
        title = elementText;
        if(title.length()>200) {
          title = title.substring(0, 200).trim();
          int nlIdx = title.indexOf('\n');
          if(nlIdx>0) title = title.substring(0, nlIdx).trim();
        } else {
          title = elementText.trim();
        }
      }
      
      if(DEBUG) {
        String textToPrint = null;
        if(elementText!=null) {
          textToPrint = elementText.trim();
          if(textToPrint.length()>100)
            textToPrint = textToPrint.substring(0, 100).trim();
        }
        System.err.println("  title: "+title);
        System.err.println("  mimetype: "+mimeType);
        System.err.println("  text: "+textToPrint);
      }
      
      if(elementText==null || elementText.trim().length()<=0) {
        // the parser failed
        System.err.println("Warning: no text parsed from element "+element);
      }
      
      if(title!=null) {
        doc.add(new Field("eltitle_"+elementID.replace(':','_'), title,
                          Field.Store.YES, Field.Index.TOKENIZED));
      }
      if(mimeType!=null) {
        doc.add(new Field("eltype_"+elementID.replace(':','_'), mimeType,
                          Field.Store.YES, Field.Index.UN_TOKENIZED));
      }
      if(elementText!=null) {
        doc.add(new Field("eltext_"+elementID.replace(':','_'), elementText,
                          Field.Store.NO, Field.Index.TOKENIZED));
      }
      String attributes[] = element.listAttributes();
      for(int i=0; attributes!=null && i<attributes.length; i++) {
        String attName = attributes[i];
        doc.add(new Field("elatt_"+elementID.replace(':','_')+"_"+attName.replace(':','_'), 
                          element.getAttribute(attName, ""),
                          Field.Store.YES, Field.Index.TOKENIZED));
      }
    } catch (Exception e) {
      System.err.println("error getting element "+element+": "+e);
      if(DEBUG) {
        e.printStackTrace(System.err);
        System.err.flush();
      }
    } finally {
      try { if(in!=null) in.close(); } catch (Throwable t) {}
    }
  }
  
  private void sendErrorResponse(OutputStream out, String msg, int code)
    throws IOException
  {
    HeaderSet response = new HeaderSet("response");
    response.addHeader("status", "error");
    if(msg!=null)
      response.addHeader("message", msg);
    if(code>=0)
      response.addHeader("code", code);
    response.writeHeaders(out);
    out.flush();
  }
  
  private void sendSuccessResponse(OutputStream out)
    throws IOException
  {
    HeaderSet response = new HeaderSet("response");
    response.addHeader("status", "success");
    response.writeHeaders(out);
    out.flush();
  }
  
  
}


